@@ -1,27 +1,27 @@
-use strict;
 package Plack::App::GitHub::WebHook;
-#ABSTRACT: GitHub WebHook receiver as Plack application
-our $VERSION = '0.4'; #VERSION
+use strict;
 use v5.10;
 use JSON qw(decode_json);
 
 use parent 'Plack::Component';
-use Plack::Util::Accessor qw(hook access app);
+use Plack::Util::Accessor qw(hook access app events safe);
 use Plack::Request;
 use Plack::Middleware::Access;
 use Carp qw(croak);
 
+our $VERSION = '0.5';
+
 sub prepare_app {
     my $self = shift;
 
-    if ($self->hook) {
-        if ( (ref $self->hook // '') eq 'CODE' ) {
-            $self->hook( [ $self->hook ] );
-        } elsif ( (ref $self->hook // '') ne 'ARRAY' ) {
-            croak "hook must be a CODEREF or ARRAYREF";
+    if ( (ref $self->hook // '') ne 'ARRAY' ) {
+        $self->hook( [ $self->hook // () ] );
+    }
+
+    foreach my $task (@{$self->hook}) {
+        if ( (ref $task // '') ne 'CODE') {
+            croak "hook must be a CODE or ARRAY of CODEs";
         }
-    } else {
-        $self->hook([]);
     }
 
     $self->access([
@@ -32,12 +32,12 @@ sub prepare_app {
 
     $self->app(
         Plack::Middleware::Access->wrap(
-            sub { $self->call_granted(shift) },
+            sub { $self->call_granted($_[0]) },
             rules => $self->access
         )
     );
 
-    $self->init;
+    $self->init; # TODO: not documented: remove?
 }
 
 sub init { }
@@ -55,8 +55,12 @@ sub call_granted {
     }
 
     my $req = Plack::Request->new($env);
-
-    my $json = eval { decode_json $req->body_parameters->{payload} };
+    my $event = $env->{'X_GITHUB_EVENT'};
+    my $json;
+    
+    if ( !$self->events or grep { $event eq $_ } @{$self->events} ) {
+        $json = eval { decode_json $req->body_parameters->{payload} };
+    }
 
     if (!$json) {
         return [400,['Content-Type'=>'text/plain','Content-Length'=>11],['Bad Request']];
@@ -72,33 +76,28 @@ sub call_granted {
 sub receive {
     my ($self, $payload) = @_;
 
-    my $ok;
     foreach my $hook (@{$self->{hook}}) {
-        return unless $hook->($payload);
-        $ok++;
+        if ($self->safe) {
+            return unless eval { $hook->($payload) } and !$@;
+        } else {
+            return unless $hook->($payload);
+        }
     } 
-    return $ok;
+
+    return scalar @{$self->{hook}};
 }
 
-
 1;
-
 __END__
 
-=pod
-
-=encoding UTF-8
-
 =head1 NAME
 
 Plack::App::GitHub::WebHook - GitHub WebHook receiver as Plack application
 
-=head1 VERSION
-
-version 0.4
-
 =head1 SYNOPSIS
 
+=head2 Basic usage
+
     use Plack::App::GitHub::WebHook;
 
     Plack::App::GitHub::WebHook->new(
@@ -108,10 +107,10 @@ version 0.4
         }
     )->to_app;
 
-=head2 Multiple hooks
+=head2 Multiple task hooks
 
-If multiple hooks are provided, they get called one by one until
-a hook returns a false value.
+A hook can consist of multiple tasks, given by an array reference. The tasks
+are called one by one until a task returns a false value.
 
     use Plack::App::GitHub::WebHook;
     use IPC::Run3;
@@ -119,6 +118,7 @@ a hook returns a false value.
     Plack::App::GitHub::WebHook->new(
         hook => [
             sub { $_[0]->{repository}{name} eq 'foo' }, # filter
+            { Filter => { repository_name => 'foo' } }, # equivalent filter
             sub { my ($payload) = @_; ...  }, # some action
             sub { run3 \@cmd ... }, # some more action
         ]
@@ -152,11 +152,39 @@ By default access is restricted to known GitHub WebHook IPs.
         }
     };
 
+=head2 Synchronize with a GitHub repository
+
+The following application automatically pulls the master branch of a GitHub
+repository into a local working directory C<$work_tree>.
+
+    use Git::Repository
+    use Plack::App::GitHub::WebHook;
+
+    Plack::App::GitHub::WebHook->new(
+        events => ['pull'],
+        safe => 1,
+        hook => [
+            sub { $_[0]->{ref} eq 'refs/heads/master' },
+            sub {
+                if ( -d "$work_tree/.git") {
+                    Git::Repository->new( work_tree => $work_tree )
+                                   ->run(qw(pull origin master));
+                } else {
+                    my $origin = $_[0]->{repository}->{clone_url};
+                    Git::Repository->run( 'clone', $origin, $work_tree );
+                }
+                1;
+            },
+            # sub { ...optional action after each pull... } 
+        ],
+    )->to_app;
+
 =head1 DESCRIPTION
 
 This L<PSGI> application receives HTTP POST requests with body parameter
-C<payload> set to a JSON object. The default use case is to receive
-L<GitHub WebHooks|https://help.github.com/articles/post-receive-hooks>.
+C<payload> set to a JSON object. The default use case is to receive 
+L<GitHub WebHooks|http://developer.github.com/webhooks/>, for instance
+L<PushEvents|http://developer.github.com/v3/activity/events/types/#pushevent>.
 
 The response of a HTTP request to this application is one of:
 
@@ -172,8 +200,8 @@ If the request was no HTTP POST.
 
 =item HTTP 400 Bad Request
 
-If the payload was no well-formed JSON. A later version of this module may add
-further validation.
+If the payload was no well-formed JSON or the C<X-GitHub-Event> header did not
+match configured events.
 
 =item HTTP 200 OK
 
@@ -193,14 +221,21 @@ This module requires at least Perl 5.10.
 
 =item hook
 
-A code reference or an array reference of code references with multiple tasks.
-Each task gets passed the encoded payload. If the task returns a true value,
-next the task is called or HTTP status code 200 is returned. Information can be
-passed from one task to the next by modifying the payload. 
+A code reference or an array of code references with tasks that are executed on
+an incoming webhook.  Each task gets passed the encoded payload. If the task
+returns a true value, next the task is called or HTTP status code 200 is
+returned. Information can be passed from one task to the next by modifying the
+payload. 
 
-If a task fails or no task was given, HTTP status code 202 is returned
-immediately. This mechanism can be used for conditional hooks or to detect
-hooks that were called successfully but failed to execute for some reason.
+If a task returns a false value or if no task was given, HTTP status code 202
+is returned immediately. This mechanism can be used for conditional hooks or to
+detect hooks that were called successfully but failed to execute for some
+reason.
+
+=item safe
+
+Wrap all hook tasks in C<< eval { ... } >> blocks to catch exceptions. A dying
+task in safe mode is equivalent to a task that returns a false value.
 
 =item access
 
@@ -209,22 +244,67 @@ for the default value. A recent list of official GitHub WebHook IPs is vailable
 at L<https://api.github.com/meta>. One should only set the access value on
 instantiation, or manually call C<prepare_app> after modification.
 
+=item events
+
+A list of L<event types|http://developer.github.com/v3/activity/events/types/>
+expected to be send with the C<X-GitHub-Event> header (e.g. C<['pull']>).
+
+=cut
+
 =back
 
 =head1 SEE ALSO
 
+=over
+
+=item
+
+GitHub WebHooks are documented at L<http://developer.github.com/webhooks/>.
+
+=item
+
 L<WWW::GitHub::PostReceiveHook> uses L<Web::Simple> to receive GitHub web
-hooks. L<Net::GitHub> and L<Pithub> provide access to GitHub APIs.
+hooks. A listener as exemplified by the module can also be created like this:
+
+    use Plack::App::GitHub::WebHook;
+    use Plack::Builder;
+    build {
+        mount '/myProject' => 
+            Plack::App::GitHub::WebHook->new(
+                hook => sub { my $payload = shift; }
+            );
+        mount '/myOtherProject' => 
+            Plack::App::GitHub::WebHook->new(
+                hook => sub { run3 \@cmd ... }
+            );
+    };
+
+=item
+
+L<Net::GitHub> and L<Pithub> provide access to GitHub APIs.
+
+=item
+
+L<App::GitHubWebhooks2Ikachan> is an application that also receives GitHub WebHooks.
+
+=back
+
+=begin markdown
+
+# STATUS
+
+[![Build Status](https://travis-ci.org/nichtich/Plack-App-GitHub-WebHook.png)](https://travis-ci.org/nichtich/Plack-App-GitHub-WebHook)
+[![Coverage Status](https://coveralls.io/repos/nichtich/Plack-App-GitHub-WebHook/badge.png?branch=master)](https://coveralls.io/r/nichtich/Plack-App-GitHub-WebHook?branch=master)
+
+=end markdown
 
 =head1 AUTHOR
 
-Jakob Voß
+Jakob Voss C<< <jakob.voss@gbv.de> >>
 
-=head1 COPYRIGHT AND LICENSE
+=head1 LICENSE
 
-This software is copyright (c) 2014 by Jakob Voß.
-
-This is free software; you can redistribute it and/or modify it under
-the same terms as the Perl 5 programming language system itself.
+This library is free software; you can redistribute it and/or modify it under
+the same terms as Perl itself.
 
 =cut
